"""Implementations for supported HSMs."""

from __future__ import annotations

import abc
from dataclasses import InitVar, dataclass
from typing import Dict, Union

import pkcs11
from cryptography.hazmat.primitives.asymmetric import dsa, ec, rsa
from cryptography.hazmat.primitives.serialization import load_der_public_key
from pkcs11.util.dsa import encode_dsa_public_key
from pkcs11.util.ec import encode_ec_public_key
from pkcs11.util.rsa import encode_rsa_public_key

import autocsr.protos.csr_pb2 as proto

PublicKey = Union[
    rsa.RSAPublicKey,
    dsa.DSAPublicKey,
    ec.EllipticCurvePublicKey,
]
HsmInfo = proto.CertificateSigningRequest.HsmInfo
KeyType = proto.CertificateSigningRequest.KeyType
HashType = proto.CertificateSigningRequest.HashType


@dataclass
class Hsm(abc.ABC):
    """Abstract class for HSMs."""

    hash_type: HashType
    key_type: KeyType

    @property
    @abc.abstractmethod
    def public_key(self) -> PublicKey:
        """Get the public key from the HSM."""

    @abc.abstractmethod
    def sign(self, message: bytes) -> bytes:
        """Sign a message with a predefined key from the HSM."""

    def pkcs11_to_crypto_key(self, pkcs11_public_key: pkcs11.KeyType) -> PublicKey:
        """Convert a pkcs11 public key to a cryptography public key."""
        if self.pkcs11_key_type == pkcs11.KeyType.RSA:
            der_public_key = encode_rsa_public_key(pkcs11_public_key)
        elif self.pkcs11_key_type == pkcs11.KeyType.DSA:
            der_public_key = encode_dsa_public_key(pkcs11_public_key)
        elif self.pkcs11_key_type == pkcs11.KeyType.EC:
            der_public_key = encode_ec_public_key(pkcs11_public_key)
        else:
            raise TypeError(f"Key type for {pkcs11_public_key} not supported")

        return load_der_public_key(der_public_key)

    @classmethod
    @abc.abstractmethod
    def from_proto(cls, hsm_info: HsmInfo, hash_info: HashType):
        """Create an HSM instance from a protobuf."""


@dataclass
class SoftHsm(Hsm):
    """An implementation of SoftHsm for signing."""

    token_label: str
    key_label: str
    user_pin: str
    so_file: str
    types: InitVar[Dict[KeyType, pkcs11.KeyType]] = {
        KeyType.RSA: pkcs11.KeyType.RSA,
    }
    hashes: InitVar[Dict[KeyType, Dict[HashType, pkcs11.Mechanism]]] = {
        KeyType.RSA: {
            HashType.SHA224: pkcs11.Mechanism.SHA224_RSA_PKCS,
            HashType.SHA256: pkcs11.Mechanism.SHA256_RSA_PKCS,
            HashType.SHA384: pkcs11.Mechanism.SHA384_RSA_PKCS,
            HashType.SHA512: pkcs11.Mechanism.SHA512_RSA_PKCS,
        },
    }

    def __post_init__(
        self,
        types: Dict[KeyType, pkcs11.KeyType],
        hashes: Dict[KeyType, Dict[HashType, pkcs11.Mechanism]],
    ):
        """Load token for SoftHSM operations."""
        lib = pkcs11.lib(self.so_file)
        self.token = lib.get_token(token_label=self.token_label)

        self.pkcs11_key_type = types[self.key_type]
        self.pkcs11_hash_type = hashes[self.key_type][self.hash_type]

    @property
    def public_key(self) -> PublicKey:
        """Get the public key from SoftHSM."""
        with self.token.open(user_pin=self.user_pin) as session:
            public_key = session.get_key(
                label=self.key_label,
                key_type=self.pkcs11_key_type,
                object_class=pkcs11.ObjectClass.PUBLIC_KEY,
            )

            return self.pkcs11_to_crypto_key(public_key)

    def sign(self, message: bytes) -> bytes:
        """Sign a message with a key from SoftHSM."""
        with self.token.open(rw=True, user_pin=self.user_pin) as session:
            private_key = session.get_key(
                label=self.key_label,
                key_type=self.pkcs11_key_type,
                object_class=pkcs11.ObjectClass.PRIVATE_KEY,
            )

            return private_key.sign(message, mechanism=self.pkcs11_hash_type)

    @classmethod
    def from_proto(cls, hsm_info: HsmInfo, hash_type: HashType):
        """Create a SoftHSM instance from a protobuf."""
        softhsm = hsm_info.softhsm

        return cls(
            hash_type=hash_type,
            key_type=hsm_info.key_type,
            token_label=softhsm.token_label,
            key_label=softhsm.key_label,
            user_pin=softhsm.user_pin,
            so_file=softhsm.so_file,
        )


class HsmFactory:
    """Create HSMs from HSMInfo configs."""

    hsms = {
        "softhsm": SoftHsm,
    }

    @staticmethod
    def from_hsm_info(hsm_info: HsmInfo, hash_type: HashType):
        """Create an HSM instance from hsm_info protobuf."""
        return HsmFactory.hsms.get(hsm_info.WhichOneof("hsm")).from_proto(
            hsm_info=hsm_info,
            hash_type=hash_type,
        )



"""Utilities for common usage."""

import json
import os
from typing import Generator, List

import yaml
from google.protobuf import json_format
from jinja2 import Template

from autocsr.protos.csr_pb2 import CertificateSigningRequest

CsrList = Generator[CertificateSigningRequest, None, None]


def load_csr(csr: dict) -> CertificateSigningRequest:
    """Load a CSR from a dictionary."""
    return json_format.Parse(json.dumps(csr), CertificateSigningRequest())


def load_csrs_from_yaml(config_file: str) -> CsrList:
    """Load a CSR from a YAML file."""
    with open(config_file, "r") as config:
        config_csrs = yaml.safe_load(config.read())

        for csr in config_csrs.values():
            yield load_csr(csr)


def load_csrs_from_jinja(config_file: str) -> CsrList:
    """Load a CSR from a Jinja template."""
    with open(config_file, "r") as config:
        config_template = Template(config.read())
        config_csrs = yaml.safe_load(config_template.render(**os.environ))

        for csr in config_csrs.values():
            yield load_csr(csr)


def load_csrs_from_file(config_file: str) -> List[CertificateSigningRequest]:
    """Load all CSRs from a file."""
    if config_file.endswith(".yaml"):
        return load_csrs_from_yaml(config_file)

    if config_file.endswith(".jinja2"):
        return load_csrs_from_jinja(config_file)

    raise TypeError(f"Config file type {config_file} is not supported")



"""Build Certificate Signing Requests."""

from __future__ import annotations

from base64 import b64decode
from dataclasses import InitVar, dataclass
from functools import partial
from typing import ClassVar, Dict, Optional, Union

from cryptography import x509
from cryptography.hazmat._types import _PRIVATE_KEY_TYPES
from cryptography.hazmat.backends.openssl.backend import Backend
from cryptography.hazmat.backends.openssl.x509 import _CertificateSigningRequest
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import dsa, ec, rsa
from cryptography.x509.oid import NameOID
from OpenSSL import crypto as openssl
from pyasn1.codec.der.decoder import decode as asn1_decode
from pyasn1.codec.der.encoder import encode as asn1_encode
from pyasn1.type.univ import BitString
from pyasn1_modules.rfc2314 import CertificationRequest

import autocsr.protos.csr_pb2 as proto
from autocsr.extensions import Extension
from autocsr.hsm import HsmFactory
from autocsr.oid import ObjectIdentifier

PrivateKey = Union[
    rsa.RSAPrivateKey,
    dsa.DSAPrivateKey,
    ec.EllipticCurvePrivateKey,
]
PublicKey = Union[
    rsa.RSAPublicKey,
    dsa.DSAPublicKey,
    ec.EllipticCurvePublicKey,
]
CsrSubject = proto.CertificateSigningRequest.Subject
HashType = proto.CertificateSigningRequest.HashType
KeyType = proto.CertificateSigningRequest.KeyType
KeyInfo = proto.CertificateSigningRequest.KeyInfo
HsmInfo = proto.CertificateSigningRequest.HsmInfo
Curve = proto.CertificateSigningRequest.Curve
GenericKeyInfo = Union[KeyInfo, HsmInfo]


class Attribute(x509.NameAttribute):
    """An x509 attribute created from configs."""

    @classmethod
    def from_field(cls, field: str, value: str):
        """Create an attribute from each field of subject info."""
        return cls(getattr(NameOID, field.upper()), value)


class Subject(x509.Name):
    """An x509 name object created from configs."""

    subject_fields = [
        "common_name",
        "country_name",
        "state_or_province_name",
        "locality_name",
        "organization_name",
        "organizational_unit_name",
        "email_address",
    ]

    @classmethod
    def from_subject(cls, subject: CsrSubject):
        """Build a x509 name object from dictionary."""
        attributes = []

        for field in cls.subject_fields:
            value = getattr(subject, field)

            if value:
                attributes.append(Attribute.from_field(field, value))

        return cls(attributes)


@dataclass
class SigningKey:
    """A signing key created from a config."""

    private_key: PrivateKey
    algorithm: hashes.HashAlgorithm = None
    hash_type: InitVar[HashType] = None
    approved_hashes: InitVar[Dict[HashType, hashes.HashAlgorithm]] = {
        HashType.SHA256: hashes.SHA256,
        HashType.SHA224: hashes.SHA224,
        HashType.SHA384: hashes.SHA384,
        HashType.SHA512: hashes.SHA512,
        HashType.SHA512_224: hashes.SHA512_224,
        HashType.SHA512_256: hashes.SHA512_256,
        HashType.BLAKE2b: partial(hashes.BLAKE2b, digest_size=64),
        HashType.BLAKE2s: partial(hashes.BLAKE2s, digest_size=32),
        HashType.SHA3_224: hashes.SHA3_224,
        HashType.SHA3_256: hashes.SHA3_256,
        HashType.SHA3_384: hashes.SHA3_384,
        HashType.SHA3_512: hashes.SHA3_512,
    }
    CURVES: ClassVar[Dict[Curve, ec.EllipticCurve]] = {
        Curve.SECP256R1: ec.SECP256R1,
        Curve.SECP384R1: ec.SECP384R1,
        Curve.SECP521R1: ec.SECP521R1,
        Curve.SECP192R1: ec.SECP192R1,
        Curve.SECP256K1: ec.SECP256K1,
        Curve.BrainpoolP256R1: ec.BrainpoolP256R1,
        Curve.BrainpoolP384R1: ec.BrainpoolP384R1,
        Curve.BrainpoolP512R1: ec.BrainpoolP512R1,
        Curve.SECT571K1: ec.SECT571K1,
        Curve.SECT409K1: ec.SECT409K1,
        Curve.SECT283K1: ec.SECT283K1,
        Curve.SECT233K1: ec.SECT233K1,
        Curve.SECT163K1: ec.SECT163K1,
        Curve.SECT571R1: ec.SECT571R1,
        Curve.SECT409R1: ec.SECT409R1,
        Curve.SECT283R1: ec.SECT283R1,
        Curve.SECT233R1: ec.SECT233R1,
        Curve.SECT163R2: ec.SECT163R2,
    }

    def __post_init__(
        self,
        hash_type: HashType,
        approved_hashes: Dict[HashType, hashes.HashAlgorithm],
    ):
        """Map HashType enums to hash algorithms."""
        self.algorithm = approved_hashes[hash_type]()

    @staticmethod
    def create_rsa_key(key_info: GenericKeyInfo) -> rsa.RSAPrivateKey:
        """Create an RSA Private Key from a key info structure."""
        if not key_info.key_size:
            key_info.key_size = 2048

        if not key_info.public_exponent:
            key_info.public_exponent = 65537

        return rsa.generate_private_key(
            public_exponent=key_info.public_exponent,
            key_size=key_info.key_size,
        )

    @staticmethod
    def create_dsa_key(key_info: GenericKeyInfo) -> dsa.DSAPrivateKey:
        """Create a DSA Private Key from a key info structure."""
        if not key_info.key_size:
            key_info.key_size = 2048

        return dsa.generate_private_key(
            key_size=key_info.key_size,
        )

    @staticmethod
    def create_ec_key(key_info: GenericKeyInfo) -> ec.EllipticCurvePrivateKey:
        """Create an EC Private Key from a key info structure."""
        return ec.generate_private_key(
            curve=SigningKey.CURVES[key_info.curve](),
        )

    @staticmethod
    def create_key(key_info: KeyInfo) -> PrivateKey:
        """Create and export a private key from a key info object."""
        if key_info.key_type == KeyType.RSA:
            private_key = SigningKey.create_rsa_key(key_info)
        elif key_info.key_type == KeyType.DSA:
            private_key = SigningKey.create_dsa_key(key_info)
        elif key_info.key_type == KeyType.EC:
            private_key = SigningKey.create_ec_key(key_info)
        else:
            raise TypeError(f"Key type: {key_info.key_type} does not exist")

        with open(key_info.key_path, "wb") as key_file:
            key_file.write(
                private_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            )

        return private_key

    @classmethod
    def from_path(cls, key_path: str, hash_type: HashType):
        """Build a Signing Key from key type and file path."""
        with open(key_path, "rb") as key_file:
            return cls(
                private_key=serialization.load_pem_private_key(
                    key_file.read(), password=None
                ),
                hash_type=hash_type,
            )

    @classmethod
    def from_key_info(cls, key_info: KeyInfo, hash_type: HashType):
        """Build a Signing Key from key information structure."""
        if key_info.create:
            return cls(
                private_key=cls.create_key(key_info),
                hash_type=hash_type,
            )

        return cls.from_path(
            key_path=key_info.key_path,
            hash_type=hash_type,
        )

    @classmethod
    def from_hsm_info(cls, hsm_info: HsmInfo, hash_type: HashType):
        """Build a Dummy Signing Key from hsm information structure."""
        if hsm_info.key_type == KeyType.RSA:
            return cls(SigningKey.create_rsa_key(hsm_info), hash_type=hash_type)

        if hsm_info.key_type == KeyType.DSA:
            return cls(SigningKey.create_dsa_key(hsm_info), hash_type=hash_type)

        if hsm_info.key_type == KeyType.EC:
            return cls(SigningKey.create_ec_key(hsm_info), hash_type=hash_type)

        raise TypeError(f"Key type: {hsm_info.key_type} does not exist")


class CertificateSigningRequest(_CertificateSigningRequest):
    """A custom certificate signing request."""

    def __repr__(self):
        """Debug representation of a CSR."""
        return (
            f"Subject: {self.subject}\n"
            f"Signature Algorithm: {self.signature_algorithm_oid}\n"
            f"Extensions: {self.extensions}\n"
            f"Valid Signature: {self.is_signature_valid}"
        )

    def set_pubkey(self, public_key: PublicKey):
        """Modify the public key of a CSR."""
        openssl_req = openssl.X509Req.from_cryptography(self)

        key_pem = public_key.public_bytes(
            serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        openssl_key = openssl.load_publickey(openssl.FILETYPE_PEM, key_pem)

        openssl_req.set_pubkey(openssl_key)

        self._x509_req = openssl_req.to_cryptography()._x509_req

    def set_signature(self, signature: bytes):
        """Return a new CSR with modified signature."""
        der_csr = self.public_bytes(serialization.Encoding.DER)
        asn1_csr, _ = asn1_decode(
            der_csr,
            asn1Spec=CertificationRequest(),
        )

        asn1_csr.setComponentByName("signature", BitString.fromOctetString(signature))

        self._x509_req = x509.load_der_x509_csr(asn1_encode(asn1_csr))._x509_req

    def export(self, filename: str):
        """Export CSR to a file."""
        with open(filename, "wb") as output_file:
            output_file.write(self.public_bytes(serialization.Encoding.PEM))


class MyBackend(Backend):
    """A custom backend to override Certificate Signing Request building."""

    def create_x509_csr(
        self,
        builder: x509.CertificateSigningRequestBuilder,
        private_key: _PRIVATE_KEY_TYPES,
        algorithm: Optional[hashes.HashAlgorithm],
    ) -> _CertificateSigningRequest:
        """Create custom Certificate Signing Requests."""
        csr = super().create_x509_csr(builder, private_key, algorithm)

        return CertificateSigningRequest(backend=self, x509_req=csr._x509_req)


class CertificateSigningRequestBuilder:
    """An x509 certificate signing request created from configs."""

    @staticmethod
    def sign_with_key_info(
        csr: proto.CertificateSigningRequest,
        builder: x509.CertificateSigningRequestBuilder,
    ):
        """Sign a CertificateSigningRequest with filesystem key."""
        key = SigningKey.from_key_info(csr.key_info, csr.hash_type)

        return builder.sign(
            private_key=key.private_key,
            algorithm=key.algorithm,
            backend=MyBackend(),
        )

    @staticmethod
    def sign_with_hsm_info(
        csr: proto.CertificateSigningRequest,
        builder: x509.CertificateSigningRequestBuilder,
    ):
        """Sign a CertificateSigningRequest with key information in an HSM."""
        hsm = HsmFactory.from_hsm_info(csr.hsm_info, csr.hash_type)

        dummy_key = SigningKey.from_hsm_info(csr.hsm_info, csr.hash_type)

        base_csr = builder.sign(
            private_key=dummy_key.private_key,
            algorithm=dummy_key.algorithm,
            backend=MyBackend(),
        )

        base_csr.set_pubkey(hsm.public_key)
        base_csr.set_signature(hsm.sign(base_csr.tbs_certrequest_bytes))

        return base_csr

    @staticmethod
    def from_csr(csr: proto.CertificateSigningRequest):
        """Build an x509 Certificate Signing Request Object from a config."""
        builder = x509.CertificateSigningRequestBuilder()
        builder = builder.subject_name(Subject.from_subject(csr.subject))

        for attribute in csr.attributes:
            builder = builder.add_attribute(
                oid=ObjectIdentifier.from_string(attribute.oid),
                value=b64decode(attribute.b64_value.encode()),
            )

        for extension in csr.extensions:
            builder = builder.add_extension(
                extval=Extension.from_proto(extension),
                critical=extension.critical,
            )

        if csr.WhichOneof("key") == "key_info":
            return CertificateSigningRequestBuilder.sign_with_key_info(csr, builder)
        elif csr.WhichOneof("key") == "hsm_info":
            return CertificateSigningRequestBuilder.sign_with_hsm_info(csr, builder)
        else:
            raise TypeError("Neither filesystem key nor HSM key selected")



"""Command line utilities for autocsr."""

import typer

from autocsr import CertificateSigningRequestBuilder, load_csrs_from_file

app = typer.Typer()


@app.command()
def create(config_file: str):
    """Create Certificate signing requests from a config file."""
    csr_list = load_csrs_from_file(config_file)

    for csr_proto in csr_list:
        csr = CertificateSigningRequestBuilder.from_csr(csr_proto)
        csr.export(csr_proto.output_path)
        typer.echo(f"Created new CSR at {csr_proto.output_path}")


def main():
    """Entrypoint to executable script."""
    app()



"""Wrappers to turn configs into x509 Extensions."""

from base64 import b64decode
from datetime import datetime
from ipaddress import ip_address
from typing import Iterable

from cryptography.x509 import extensions, general_name
from cryptography.x509 import name as x509_name

import autocsr.protos.csr_pb2 as proto
from autocsr.oid import ObjectIdentifier

DATE_FMT = "%d/%m/%Y %H:%M:%S"

CsrExtension = proto.CertificateSigningRequest.Extension


class ExtensionType:
    """Factory for returning a parameter-less extension from a protobuf."""

    @staticmethod
    def from_proto(extension: CsrExtension) -> extensions.ExtensionType:
        """Return an extension instance based on enum extension type."""
        this_extension = extension.extension_type

        if this_extension == extension.ExtensionType.OCSPNoCheck:
            return extensions.OCSPNoCheck()

        if this_extension == extension.ExtensionType.PrecertPoison:
            return extensions.PrecertPoison()

        raise TypeError(f"No known extension: {this_extension}")


# Unsupported by cryptography
class CRLNumber(extensions.CRLNumber):
    """Wrapper for CRLNumber Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a CRLNumber extension from a protobuf."""
        this_extension = extension.crl_number

        return cls(
            crl_number=this_extension.crl_number,
        )


class SubjectKeyIdentifier(extensions.SubjectKeyIdentifier):
    """Wrapper for SubjectKeyIdentifier Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a SubjectKeyIdentifier extension from a protobuf."""
        this_extension = extension.subject_key_identifier

        return cls(digest=b64decode(this_extension.b64_digest.encode()))


class BasicConstraints(extensions.BasicConstraints):
    """Wrapper for BasicConstraints Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a BasicConstraints extension from a protobuf."""
        this_extension = extension.basic_constraints

        path_length = this_extension.path_length

        if not path_length:
            path_length = None

        return cls(
            ca=this_extension.ca,
            path_length=path_length,
        )


# Unsupported by cryptography
class DeltaCRLIndicator(extensions.DeltaCRLIndicator):
    """Wrapper for DeltaCRLIndicator Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a DeltaCRLIndicator extension from a protobuf."""
        this_extension = extension.delta_crl_indicator

        return cls(crl_number=this_extension.crl_number)


class PolicyConstraints(extensions.PolicyConstraints):
    """Wrapper for PolicyConstraints Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a PolicyConstraints Extension from a protobuf."""
        this_extension = extension.policy_constraints

        require_explicit_policy = this_extension.require_explicit_policy
        inhibit_policy_mapping = this_extension.inhibit_policy_mapping

        if not require_explicit_policy:
            require_explicit_policy = None

        if not inhibit_policy_mapping:
            inhibit_policy_mapping = None

        return cls(
            require_explicit_policy=require_explicit_policy,
            inhibit_policy_mapping=inhibit_policy_mapping,
        )


class NoticeReference(extensions.NoticeReference):
    """Wrapper for NoticeReference from config file."""

    @classmethod
    def from_proto(cls, notice: CsrExtension.NoticeReference):
        """Create a Noticereference from a protobuf."""
        organization = notice.organization

        if not organization:
            organization = None

        return cls(
            organization=organization,
            notice_numbers=notice.notice_numbers,
        )


class UserNotice(extensions.UserNotice):
    """Wrapper for UserNotice from config file."""

    @classmethod
    def from_proto(cls, notices: Iterable[CsrExtension.UserNotice]):
        """Create a list of UserNotice form a protobuf."""
        information_list = []

        for notice in notices:
            notice_reference = NoticeReference.from_proto(notice.notice_reference)
            explicit_text = notice.explicit_text

            if not notice_reference:
                notice_reference = None

            if not explicit_text:
                explicit_text = None

            information_list.append(
                cls(
                    notice_reference=notice_reference,
                    explicit_text=explicit_text,
                )
            )

        return information_list


class PolicyInformation(extensions.PolicyInformation):
    """Wrapper for PolicyInformation from config file."""

    @classmethod
    def from_proto(cls, policies: Iterable[CsrExtension.PolicyInformation]):
        """Create a list of PolicyInformation from a protobuf."""
        information_list = []

        for policy in policies:
            policy_qualifiers = []

            for string in policy.string_qualifiers:
                policy_qualifiers.append(string)

            for user_notice in UserNotice.from_proto(policy.user_qualifiers):
                policy_qualifiers.append(user_notice)

            information_list.append(
                cls(
                    policy_identifier=ObjectIdentifier.from_string(
                        policy.policy_identifier
                    ),
                    policy_qualifiers=policy_qualifiers,
                )
            )

        return information_list


class CertificatePolicies(extensions.CertificatePolicies):
    """Wrapper for CertificatePolicies Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a CertificatePolicies extension from a protobuf."""
        this_extension = extension.certificate_policies

        return cls(policies=PolicyInformation.from_proto(this_extension.policies))


class ExtendedKeyUsage(extensions.ExtendedKeyUsage):
    """Wrapper for ExtendedKeyUsage extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a ExtendedKeyUsage extension from a protobuf."""
        this_extension = extension.extended_key_usage

        return cls(
            usages=(ObjectIdentifier.from_string(oid) for oid in this_extension.usages),
        )


class TLSFeature(extensions.TLSFeature):
    """Wrapper for TLSFeature Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a TLSFeature extension from a protobuf."""
        this_extension = extension.tls_feature

        return cls(
            features=(
                extensions.TLSFeatureType(feature)
                for feature in this_extension.features
            )
        )


class InhibitAnyPolicy(extensions.InhibitAnyPolicy):
    """Wrapper for InhibitAnyPolicy Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a InhibitAnyPolicy extension from a protobuf."""
        this_extension = extension.inhibit_any_policy

        return cls(
            skip_certs=this_extension.skip_certs,
        )


class KeyUsage(extensions.KeyUsage):
    """Wrapper for KeyUsage Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a KeyUsage extension from a protobuf."""
        this_extension = extension.key_usage

        return cls(
            digital_signature=this_extension.digital_signature,
            content_commitment=this_extension.content_commitment,
            key_encipherment=this_extension.key_encipherment,
            data_encipherment=this_extension.data_encipherment,
            key_agreement=this_extension.key_agreement,
            key_cert_sign=this_extension.key_cert_sign,
            crl_sign=this_extension.crl_sign,
            encipher_only=this_extension.encipher_only,
            decipher_only=this_extension.decipher_only,
        )


class ReasonFlags:
    """Map int enum ReasonFlags to exceptions.ReasonFlag enums."""

    ProtoFlag = CsrExtension.ReasonFlags
    ExtFlag = extensions.ReasonFlags

    flags = {
        ProtoFlag.unspecified: ExtFlag.unspecified,
        ProtoFlag.key_compromise: ExtFlag.key_compromise,
        ProtoFlag.ca_compromise: ExtFlag.ca_compromise,
        ProtoFlag.affiliation_changed: ExtFlag.affiliation_changed,
        ProtoFlag.superseded: ExtFlag.superseded,
        ProtoFlag.cessation_of_operation: ExtFlag.cessation_of_operation,
        ProtoFlag.privilege_withdrawn: ExtFlag.privilege_withdrawn,
        ProtoFlag.aa_compromise: ExtFlag.aa_compromise,
        ProtoFlag.remove_from_crl: ExtFlag.remove_from_crl,
    }

    @staticmethod
    def from_proto(reason: ProtoFlag) -> ExtFlag:
        """Convert a protobuf version ReasonFlag into the extensions version."""
        return ReasonFlags.flags[reason]


# Unsupported by cryptography
class CRLReason(extensions.CRLReason):
    """Wrapper for CRLReason Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a CRLReason extension from a protobuf."""
        this_extension = extension.crl_reason

        return cls(
            reason=ReasonFlags.from_proto(this_extension.reason),
        )


# Unsupported by cryptography
class InvalidityDate(extensions.InvalidityDate):
    """Wrapper for InvalidityDate Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a InvalidityDate from a protobuf."""
        this_extension = extension.invalidity_date

        return cls(
            invalidity_date=datetime.strptime(this_extension.invalidity_date, DATE_FMT),
        )


# Currently unsupported by autocsr
class PrecertificateSignedCertificateTimestamps(
    extensions.PrecertificateSignedCertificateTimestamps
):
    """
    Wrapper for PrecertificateSignedCertificateTimestamps.

    Wrapper for PrecertificateSignedCertificateTimestamps Extension from config
    file.
    """

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """
        Create a PrecertificateSignedCertificateTimestamps.

        Create a PrecertificateSignedCertificateTimestamps from a protobuf.
        """
        raise TypeError(
            "autocsr currently does not support signed certificate timestamps"
        )


# Currently unsupported by autocsr
class SignedCertificateTimestamps(extensions.SignedCertificateTimestamps):
    """Wrapper for SignedCertificateTimestamps Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a SignedCertificateTimestamps from a protobuf."""
        raise TypeError(
            "autocsr currently does not support signed certificate timestamps"
        )


# Unsupported by cryptography
class OCSPNonce(extensions.OCSPNonce):
    """Wrapper for OCSPNonce Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a OCSPNonce from a protobuf."""
        this_extension = extension.ocsp_nonce

        return cls(nonce=b64decode(this_extension.b64_nonce.encode()))


class RFC822Name(general_name.RFC822Name):
    """Wrapper for RFC822Name general name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.GeneralName):
        """Create a RFC822Name from a protobuf."""
        this_name = name.rfc_822_name

        return cls(value=this_name.value)


class DNSName(general_name.DNSName):
    """Wrapper for DNSName general name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.GeneralName):
        """Create a DNSName from a protobuf."""
        this_name = name.dns_name

        return cls(value=this_name.value)


class UniformResourceIdentifier(general_name.UniformResourceIdentifier):
    """Wrapper for UniformResourceIdentifier general name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.GeneralName):
        """Create a UniformResourceIdentifier from a protobuf."""
        this_name = name.uniform_resource_identifier

        return cls(value=this_name.value)


class NameAttribute(x509_name.NameAttribute):
    """Wrapper for NameAttribute from config file."""

    @classmethod
    def from_proto(cls, attribute: CsrExtension.NameAttribute):
        """Create a NameAttribute from a config file."""
        return cls(
            oid=ObjectIdentifier.from_string(attribute.oid),
            value=attribute.value,
        )


class Name(x509_name.Name):
    """Wrapper for Name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.Name):
        """Create a Name from a config file."""
        return cls(
            attributes=(
                NameAttribute.from_proto(attribute) for attribute in name.attributes
            ),
        )


class RelativeDistinguishedName(x509_name.RelativeDistinguishedName):
    """Wrapper for RelativeDistinguishedName from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.Name):
        """Create a Name from a config file."""
        return cls(
            attributes=(
                NameAttribute.from_proto(attribute) for attribute in name.attributes
            ),
        )


class DirectoryName(general_name.DirectoryName):
    """Wrapper for DirectoryName general name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.GeneralName):
        """Create a DirectoryName from a protobuf."""
        this_name = name.directory_name

        return cls(value=Name.from_proto(this_name.value))


class RegisteredID(general_name.RegisteredID):
    """Wrapper for RegisteredID general name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.GeneralName):
        """Create a RegisteredID from a protobuf."""
        this_name = name.registered_id

        return cls(value=ObjectIdentifier.from_string(this_name.oid))


class IPAddress(general_name.IPAddress):
    """Wrapper for an IPAddress general name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.GeneralName):
        """Create a IPAddress from a protobuf."""
        this_name = name.ip_address

        return cls(value=ip_address(this_name.value))


# Untested
class OtherName(general_name.OtherName):
    """Wrapper for an OtherName general name from config file."""

    @classmethod
    def from_proto(cls, name: CsrExtension.GeneralName):
        """Create a OtherName from a protobuf."""
        this_name = name.other_name

        return cls(
            type_id=ObjectIdentifier.from_string(this_name.oid),
            value=b64decode(this_name.b64_value.encode()),
        )


class GeneralName:
    """Factory for creating General Names from a config."""

    name_types = {
        "rfc_822_name": RFC822Name,
        "dns_name": DNSName,
        "uniform_resource_identifier": UniformResourceIdentifier,
        "directory_name": DirectoryName,
        "registered_id": RegisteredID,
        "ip_address": IPAddress,
        "other_name": OtherName,
    }

    @staticmethod
    def from_proto(name: CsrExtension.GeneralName) -> general_name.GeneralName:
        """Create a GeneralName instance from a protobuf."""
        return GeneralName.name_types.get(name.WhichOneof("name")).from_proto(name)


class AuthorityKeyIdentifier(extensions.AuthorityKeyIdentifier):
    """Wrapper for AuthorityKeyIdentifier Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a AuthorityKeyIdentifier from a protobuf."""
        this_extension = extension.authority_key_identifier

        key_identifier = this_extension.key_identifier
        authority_cert_serial_number = this_extension.authority_cert_serial_number

        if not key_identifier:
            key_identifier = None

        if not authority_cert_serial_number:
            authority_cert_serial_number = None

        return cls(
            key_identifier=key_identifier,
            authority_cert_issuer=(
                GeneralName.from_proto(name)
                for name in this_extension.authority_cert_issuer
            ),
            authority_cert_serial_number=authority_cert_serial_number,
        )


class AccessDescription(extensions.AccessDescription):
    """Wrapper for AccessDescription from a config file."""

    @classmethod
    def from_proto(cls, description: CsrExtension.AccessDescription):
        """Create a AccessDescription from a protobuf."""
        return cls(
            access_method=ObjectIdentifier.from_string(description.access_method),
            access_location=GeneralName.from_proto(description.access_location),
        )


class AuthorityInformationAccess(extensions.AuthorityInformationAccess):
    """Wrapper for AuthorityInformationAccess Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a AuthorityInformationAccess from a protobuf."""
        this_extension = extension.authority_information_access

        return cls(
            descriptions=(
                AccessDescription.from_proto(description)
                for description in this_extension.descriptions
            )
        )


class SubjectInformationAccess(extensions.SubjectInformationAccess):
    """Wrapper for SubjectInformationAccess Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a SubjectInformationAccess from a protobuf."""
        this_extension = extension.subject_information_access

        return cls(
            descriptions=(
                AccessDescription.from_proto(description)
                for description in this_extension.descriptions
            )
        )


class DistributionPoint(extensions.DistributionPoint):
    """Wrapper for DistributionPoint from config file."""

    @classmethod
    def from_proto(cls, dist: CsrExtension.DistributionPoint):
        """Create a DistributionPoint from a protobuf."""
        full_name = [GeneralName.from_proto(name) for name in dist.full_name]
        relative_name = RelativeDistinguishedName.from_proto(dist.relative_name)
        reasons = frozenset((ReasonFlags.from_proto(reason) for reason in dist.reasons))
        crl_issuer = [GeneralName.from_proto(name) for name in dist.crl_issuer]

        if not full_name:
            full_name = None

        if not relative_name:
            relative_name = None

        if not reasons:
            reasons = None

        if not crl_issuer:
            crl_issuer = None

        return cls(
            full_name=full_name,
            relative_name=relative_name,
            reasons=reasons,
            crl_issuer=crl_issuer,
        )


class CRLDistributionPoints(extensions.CRLDistributionPoints):
    """Wrapper for CRLDistributionPoints Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a CRLDistributionPoints from a protobuf."""
        this_extension = extension.crl_distribution_points

        return cls(
            distribution_points=(
                DistributionPoint.from_proto(dist)
                for dist in this_extension.distribution_points
            )
        )


class FreshestCRL(extensions.FreshestCRL):
    """Wrapper for FreshestCRL Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a FreshestCRL from a protobuf."""
        this_extension = extension.freshest_crl

        return cls(
            distribution_points=(
                DistributionPoint.from_proto(dist)
                for dist in this_extension.distribution_points
            )
        )


class NameConstraints(extensions.NameConstraints):
    """Wrapper for a NameConstraints extension from a config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a NameConstraints from a protobuf."""
        this_extension = extension.name_constraints

        return cls(
            permitted_subtrees=(
                GeneralName.from_proto(name)
                for name in this_extension.permitted_subtrees
            ),
            excluded_subtrees=(
                GeneralName.from_proto(name)
                for name in this_extension.excluded_subtrees
            ),
        )


class SubjectAlternativeName(extensions.SubjectAlternativeName):
    """Wrapper for SubjectAlternativeName Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a SubjectAlternativeName from a protobuf."""
        this_extension = extension.subject_alternative_name

        return cls(
            general_names=(
                GeneralName.from_proto(name) for name in this_extension.general_names
            ),
        )


class IssuerAlternativeName(extensions.IssuerAlternativeName):
    """Wrapper for IssuerAlternativeName Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a IssuerAlternativeName from a protobuf."""
        this_extension = extension.issuer_alternative_name

        return cls(
            general_names=(
                GeneralName.from_proto(name) for name in this_extension.general_names
            ),
        )


# Unsupported by cryptography
class CertificateIssuer(extensions.CertificateIssuer):
    """Wrapper for CertificateIssuer Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a CertificateIssuer from a protobuf."""
        this_extension = extension.certificate_issuer

        return cls(
            general_names=(
                GeneralName.from_proto(name) for name in this_extension.general_names
            ),
        )


# Unsupported by cryptography
class IssuingDistributionPoint(extensions.IssuingDistributionPoint):
    """Wrapper for IssuingDistributionPoint Extension from config file."""

    @classmethod
    def from_proto(cls, extension: CsrExtension):
        """Create a IssuingDistributionPoint from a protobuf."""
        this_extension = extension.issuing_distribution_point

        full_name = [GeneralName.from_proto(name) for name in this_extension.full_name]
        relative_name = RelativeDistinguishedName.from_proto(
            this_extension.relative_name
        )
        reasons = frozenset(
            (
                ReasonFlags.from_proto(reason)
                for reason in this_extension.only_some_reasons
            )
        )

        if not full_name:
            full_name = None

        if not relative_name:
            relative_name = None

        if not reasons:
            reasons = None

        return cls(
            full_name=full_name,
            relative_name=relative_name,
            only_contains_user_certs=this_extension.only_contains_user_certs,
            only_contains_ca_certs=this_extension.only_contains_ca_certs,
            only_some_reasons=reasons,
            indirect_crl=this_extension.indirect_crl,
            only_contains_attribute_certs=this_extension.only_contains_attribute_certs,
        )


class Extension:
    """A factory for creating x509 Extensions from config."""

    extension_list = {
        "extension_type": ExtensionType,
        # "crl_number": CRLNumber,
        "subject_key_identifier": SubjectKeyIdentifier,
        "basic_constraints": BasicConstraints,
        # "delta_crl_indicator": DeltaCRLIndicator,
        "policy_constraints": PolicyConstraints,
        "certificate_policies": CertificatePolicies,
        "extended_key_usage": ExtendedKeyUsage,
        "tls_feature": TLSFeature,
        "inhibit_any_policy": InhibitAnyPolicy,
        "key_usage": KeyUsage,
        # "crl_reason": CRLReason,
        # "invalidity_date": InvalidityDate,
        # "precertificate_signed_certificate_timestamps": PrecertificateSignedCertificateTimestamps,
        # 'signed_certificate_timestamps': SignedCertificateTimestamps,
        # "ocsp_nonce": OCSPNonce,
        "authority_key_identifier": AuthorityKeyIdentifier,
        "authority_information_access": AuthorityInformationAccess,
        "subject_information_access": SubjectInformationAccess,
        "crl_distribution_points": CRLDistributionPoints,
        "freshest_crl": FreshestCRL,
        "name_constraints": NameConstraints,
        "subject_alternative_name": SubjectAlternativeName,
        "issuer_alternative_name": IssuerAlternativeName,
        # "certificate_issuer": CertificateIssuer,
        # "issuing_distribution_point": IssuingDistributionPoint,
    }

    @staticmethod
    def get_extension_type(extension: CsrExtension) -> extensions.ExtensionType:
        """Get the custom extension wrapper for an extension protobuf."""
        return Extension.extension_list.get(extension.WhichOneof("extension"))

    @staticmethod
    def from_proto(extension: CsrExtension) -> extensions.ExtensionType:
        """Create extensions from an extension proto."""
        extension_type = Extension.get_extension_type(extension)

        return extension_type.from_proto(extension)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autocsr', 'autocsr.protos']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.3,<3.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'cryptography>=3.4.7,<4.0.0',
 'protobuf>=3.15.8,<4.0.0',
 'pyOpenSSL>=20.0.1,<21.0.0',
 'pyasn1-modules>=0.2.8,<0.3.0',
 'python-pkcs11>=0.7.0,<0.8.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['autocsr = autocsr.cli:main']}

setup_kwargs = {
    'name': 'autocsr',
    'version': '1.0.6',
    'description': '',
    'long_description': None,
    'author': 'Max Wolfe',
    'author_email': 'max@securitywolfe.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
